package BK::Server;

=head1 NAME

BK::Server - class to support the server side of BK

=head1 SYNOPSIS

     require BK::Server;

=head1 DESCRIPTION

=cut

require 5.002;
use BK::Composer;
use POSIX;

sub new {
    my ($proto, $libdir, $bookhome) = @_;
    my $class = ref($proto) || $proto;
    my $self  = {
        HOME        => $bookhome,
        LIBDIR      => $libdir,
	FD          => undef,
        _RELEASED   => 0,
	COMMANDS    => undef,
    };
    bless ($self, $class);

    $self->bootstrapCommands();
    return $self;
}

sub home {
    my $self = shift;
    if (@_) { $self->{HOME} = shift }
    return $self->{HOME};
}

sub bindir {
    my $self = shift;
    if (@_) { $self->{BINDIR} = shift }
    return $self->{BINDIR};
}
sub libdir {
    my $self = shift;
    if (@_) { $self->{LIBDIR} = shift }
    return $self->{LIBDIR};
}

sub fd {
    my $self = shift;
    if (@_) {
        $self->{FD} = shift;
        if (defined $self->{FD}) {
            my $old  = select($self->{FD});
            $| = 1;
            select($old);
        }
    }
    return $self->{FD};
}

sub addCommand {
    my ($self, $cmdName, $cmdHash) = @_;

    # print "adding command: $cmdName\n";

    $self->{COMMANDS}->{$cmdName} = $cmdHash;
}

sub run {
    my $self = shift;
    
    while (!$self->{_RELEASED}) {
        eval {
            my ($table, $args) = &ungenerate($self->fd());
            $self->respond($table, $args);
        };
        if ($@) {
	    # Output an error and try to send one to the client.
            local $SIG{PIPE} = 'IGNORE';

            print "ERROR: ", $@;
            $@ =~ s/ at.*\n?$//,
            print {$self->fd()} &generate({
                     data => $@,
                     func => 'client-OutputError',
               });
            $self->release();
            $self->{_RELEASED} = 1;
        }
    }

    $self->fd(undef);
    $self->{_RELEASED} = 0;
}

sub respond {
    my ($self, $table, $args) = @_;
    
    (my $cmd = $self->{COMMANDS}->{$table->{'func'}});
    die "Unknown command.  Try 'bk help' for info." if (not defined $cmd);
    &{$cmd->{'func'}}($self, $table, $args);
}

sub bootstrapCommands {
    my $self = shift;
    my $libdir = $self->{LIBDIR} . "/BK/ServFunc";

    opendir(LIBDIR, $libdir) || die "Can't open $libdir: $!";
    foreach $lib (map { chop;chop;chop; "BK::ServFunc::$_"; }
			grep {/.*\.pm$/} readdir(LIBDIR)) {
	# print "found $lib\n";

	my $cmds = '';
	my $home = $self->home();
        # Untaint $home.
        $home =~ /(.*)/; $home = $1;

	eval "require $lib"; warn $@ if $@;
	eval "\$cmds = &${lib}::register('$home')";
	if ($@) {
	    warn $@;
	} else {
	    # print "adding functions in $lib:\n";
	    while (($key, $val) = each %$cmds) {
		# print "$key, $val\n";
		$self->addCommand($key, $val);
	    }
	}
    }
    closedir(LIBDIR);
}


# Generic functions

sub release {
    my $self = shift;
    print {$self->fd()} &generate({
                     func   => 'release',
               });
}

sub editEntry {
    my ($self, $name, $ID, $tID) = @_;
   
    print {$self->fd()} &generate({
                     data => $self->generateTemplate($name, $ID, $tID),
                     confirm => "server-Update" . $name,
                     decline => 'server-ForgetIt',
                     func => 'client-EditData',
               });
}

sub readEntry {
    my ($self, $name) = @_;
   
    print {$self->fd()} &generate({
                     confirm => "server-Update" . $name,
                     decline => 'server-ForgetIt',
                     func => 'client-ReadData',
               });
}

sub writeEntry {
    my ($self, $name, $ID, $tID) = @_;

    print {$self->fd()} &generate({
                     data => $self->generateTemplate($name, $ID, $tID),
                     func => 'client-OutputData',
               });
    $self->release();
}

sub deleteEntry {
    my ($self, $name, $ID) = @_;

    $self->reallyDelete($name, $ID);

    print {$self->fd()} &generate({
                     data => "$name '$ID' deleted.",
                     func => 'client-OutputInfo',
               });
    $self->release();
}

sub generateTemplate {
    my ($self, $name, $ID, $tID) = @_;
   
    # Find header, description and entry (if exists)
    my ($f, $t, $d) = $self->getHeader($name);
    my @fields = @$f;
    my @desc = @$d;
    my @entry = ();

    if (not defined $tID) {
        @entry = $self->getEntries($name, "$name eq '$ID'");
    } else {
        @entry = $self->getEntries($name, "$name eq '$tID'");
	die "No existing $name named \"$tID\""
		if (!@entry);
    }

    if (!@entry) {
	@entry = $ID;
    } else {
	die "${name}.rdb inconsisent: duplicate ID \"$ID\""
		if (@entry > 1);
	@entry = split(/\t/, $entry[0]);
        $entry[0] = $ID if (defined $tID);
    }
    my $ln = 0;
    foreach (@fields) {
        $ln = length if (length > $ln);
    }
    $ln++;

    $header = "# A BANAL $name Specification.\n#\n";
    $content = "\n";
    foreach $desc (@desc) {
        $field = shift(@fields);
        $val = shift(@entry);
        $header .= "# " . pack("A$ln", $field . ":") . " $desc\n";
        $content .= pack("A$ln", $field . ":") . " ";
        if (defined($val)) {
	    $val =~ s/\\N/\n/g;
	    $val =~ s/\\T/\t/g;
            $content .= $val;
        } else {
            $content .= '';
        }
        $content .= "\n\n";
    }

    return $header . $content;
}

sub reallyDelete {
    my ($self, $name, $ID) = @_;
   
    my $db = $self->home() . "/" . $name . ".rdb";
    open(DATABASE, "<$db") || die "Can't open $db: $!\n";
    my $tmpfile = $db . "." . $$;
    
    # Untaint $db and $tmpfile.
    $db =~ /(.*)/; $db = $1;
    $tmpfile =~ /(.*)/; $tmpfile = $1;

    open(TMPFILE, ">$tmpfile") || die "Can't open $tmpfile: $!\n";
   
    while (<DATABASE>) {
        if (!/^\s*#/) {                 # not a comment
            if (++$ln > 2) {            # not the header
                @F = split( /\t/, $_);
                next if ($F[0] eq $ID); # too bad, so sad, bye-bye
            }
        }
        print TMPFILE;
    }

    close TMPFILE;
    close DATABASE;
    rename $tmpfile, $db;
}

sub newID {
    my ($self, $name) = @_;
    (my $baseID = substr($name, 0, 3)) =~ tr/A-Z/a-z/;
 
    my $db = $self->home() . "/" . $name . ".rdb";
    open(DATABASE, "<$db") || die "Can't open $db: $!\n";

    $num = 0;   
    while (<DATABASE>) {
        if (!/^\s*#/) {                 # not a comment
            if (++$ln > 2) {            # not the header
                @F = split( /\t/, $_);
                if ($F[0] =~ /^$baseID(\d+)$/ && $1 > $num) { # got a match
                    $num = $1;
                }
            }
        }
    }
    close DATABASE;

    $num++;
    $baseID . sprintf("%6.6u", $num);
}

sub calcTaxes {
    my ($self, $val, $taxCode) = @_;
    my %taxes = ();

    foreach (split '/', $taxCode) {
	my @tax = $self->getEntries('Category',
				    "Field eq Tax and Value eq '$_'",
				    'TaxRate');
        $taxes{$_} = $tax[0] *$val;
    }
    %taxes;
}

sub getHeader {
    my ($self, $name) = @_;
   
    my $db = $self->home() . "/" . $name . ".rdb";
    open(DATABASE, "<$db") || die "Can't open $db: $!\n";
   
    # Find header, description and entry (if exists)
    my $foundFields = 0;
    my (@fields, @types, @desc);

    while (<DATABASE>) {
        next if /^\s*#/;   # comment;
        chop;
        if ($foundFields) {# This is the decription
	    @types = map { s/\s.*//g; $_; } split(/\t/);
	    @desc = map { s/\S*\s*//; $_; } split(/\t/);
            last;
        }
        @fields = split(/\t/);
	$foundFields = 1;
    }

    close DATABASE;

    \(@fields, @types, @desc);
}

sub getEntries {
    my ($self, $name, $args, $fields, $keepHead) = @_;
    my ($columns, $stripheader) = ('', '');
    my ($ROW, $COL, $HEAD) =
		map { $self->bindir() . "/$_"; } qw{row column headchg};
   
    $args =~ s/([();?*|&<>])/\\$1/g;
    if (defined $fields && length($fields)) {
        $columns = "| $COL $fields";
    }
    $keepHead = 0 if (not defined($keepHead));
    $stripheader = "| $HEAD -del" if (!$keepHead);

    my $db = $self->home() . "/" . $name . ".rdb";

    # Untaint $db, $args and $columns.
    $db =~ /(.*)/; $db = $1;
    $args =~ /(.*)/; $args = $1;
    $columns =~ /(.*)/; $columns = $1;
    $stripheader =~ /(.*)/; $stripheader = $1;

    my @entries = qx{ $ROW $args < $db $columns $stripheader };
    chop @entries;
    @entries;
}

sub verify {
    my ($self, $val, $field, $type, $desc) = @_;
    my ($isReq, $isDate, $isCat, $isNum, $ID) = (0, 0, 0, 0, undef);

    # Find out properties of this field.

    $isNum = 1 if ($type =~ /^\d*N$/);
    $ID = $1   if ($desc =~ /^([A-Z][a-z]*) ID/);

    ($desc, $_) = split('@@', $desc);
    if (defined) {
        $isReq = 1 if (/REQUIRED/);
        $isCat = 1 if (/CATEGORIZED/);
        $isDate= 1 if (/YYYY.MMDD/);
    }

    # Now verify the data.

    if (not length $val) {
        die "\"$field\" is a required field." if $isReq;
        return;
    }

    if ($isDate) {
        die "\"$val\" is not a valid date in \"$field\" field."
                if (!($val =~ /^(\d{4})\.(\d{2})(\d{2})$/));
        my ($y, $m, $d) = ($1, $2, $3);

        die "bad month \"$m\"." if ($m < 1 || $m > 12);
        die "bad day \"$d\"."   if ($d < 1 || $d > 31);
        die "bad year \"$y\"."  if ($y < 1900);
    }

    if ($isNum) {
        die "\"$val\" is not a valid number or date in \"$field\" field."
                if (!($val =~ /^(\d+\.?\d*|\.\d+)$/));
    }

    if ($isCat) {
        my @e = $self->getEntries('Category',
                                  "Field eq '$field' and Value eq '$val'");
        die "\"$val\" is not a valid value for \"$field\"." if (not @e);
    }

    if (defined $ID) {
        my @e = $self->getEntries($ID,
                                  "$ID eq '$val'");
        die "\"$val\" is not a valid $ID." if (not @e);
    }
}

1;
